import unittest
from io import StringIO

from requests.exceptions import HTTPError
from requests_mock import Mocker

from coinbase.__version__ import __version__
from coinbase.rest import RESTClient
from tests.constants import TEST_API_KEY, TEST_API_SECRET


class RestBaseTest(unittest.TestCase):
    def test_no_api_key(self):
        with self.assertRaises(Exception):
            RESTClient(None, None)

        with self.assertRaises(Exception):
            RESTClient("test_key", None)

    def test_get(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/accounts",
                json=expected_response,
            )

            params = {"limit": 2}
            kwargs = {"test_kwarg": "test"}
            accounts = client.get("/api/v3/brokerage/accounts", params, **kwargs)

            captured_request = m.request_history[0]
            captured_query = captured_request.query
            captured_headers = captured_request.headers

            self.assertEqual(captured_request.method, "GET")

            self.assertEqual(captured_query, "limit=2&test_kwarg=test")

            self.assertTrue("User-Agent" in captured_headers)
            self.assertEqual(
                captured_headers["User-Agent"], "coinbase-advanced-py/" + __version__
            )
            self.assertTrue("Authorization" in captured_headers)
            self.assertTrue(captured_headers["Authorization"].startswith("Bearer "))

            self.assertEqual(accounts, expected_response)

    def test_post(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/portfolios",
                json=expected_response,
            )

            data = {"name": "TestName"}
            kwargs = {"test_kwarg": "test"}

            portfolio = client.post("/api/v3/brokerage/portfolios", data=data, **kwargs)

            captured_request = m.request_history[0]
            captured_json = captured_request.json()
            captured_headers = captured_request.headers

            self.assertEqual(captured_request.method, "POST")

            self.assertEqual(captured_json, {"name": "TestName", "test_kwarg": "test"})

            self.assertTrue("User-Agent" in captured_headers)
            self.assertEqual(
                captured_headers["User-Agent"], "coinbase-advanced-py/" + __version__
            )
            self.assertTrue("Authorization" in captured_headers)
            self.assertTrue(captured_headers["Authorization"].startswith("Bearer "))

            self.assertEqual(portfolio, expected_response)

    def test_put(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "PUT",
                "https://api.coinbase.com/api/v3/brokerage/portfolios/1234",
                json=expected_response,
            )

            data = {"name": "TestName"}
            kwargs = {"test_kwarg": "test"}

            portfolio = client.put(
                "/api/v3/brokerage/portfolios/1234", data=data, **kwargs
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()
            captured_headers = captured_request.headers

            self.assertEqual(captured_request.method, "PUT")

            self.assertEqual(captured_json, {"name": "TestName", "test_kwarg": "test"})

            self.assertTrue("User-Agent" in captured_headers)
            self.assertEqual(
                captured_headers["User-Agent"], "coinbase-advanced-py/" + __version__
            )
            self.assertTrue("Authorization" in captured_headers)
            self.assertTrue(captured_headers["Authorization"].startswith("Bearer "))

            self.assertEqual(portfolio, expected_response)

    def test_delete(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "DELETE",
                "https://api.coinbase.com/api/v3/brokerage/portfolios/1234",
                json=expected_response,
            )

            kwargs = {"test_kwarg": "test"}

            portfolio = client.delete("/api/v3/brokerage/portfolios/1234", **kwargs)

            captured_request = m.request_history[0]
            captured_headers = captured_request.headers

            self.assertEqual(captured_request.method, "DELETE")

            self.assertEqual(captured_request.json(), kwargs)

            self.assertTrue("User-Agent" in captured_headers)
            self.assertEqual(
                captured_headers["User-Agent"], "coinbase-advanced-py/" + __version__
            )
            self.assertTrue("Authorization" in captured_headers)
            self.assertTrue(captured_headers["Authorization"].startswith("Bearer "))

            self.assertEqual(portfolio, expected_response)

    def test_client_error(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/accounts",
                status_code=400,
            )

            with self.assertRaises(HTTPError):
                client.get("/api/v3/brokerage/accounts")

    def test_server_error(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/accounts",
                status_code=500,
            )

            with self.assertRaises(HTTPError):
                client.get("/api/v3/brokerage/accounts")

    def test_key_file_string(self):
        try:
            RESTClient(key_file="tests/test_api_key.json")
        except Exception as e:
            self.fail(f"An unexpected exception occurred: {e}")

    def test_key_file_object(self):
        try:
            key_file_object = StringIO(
                '{"name": "test-api-key-name","privateKey": "test-api-key-private-key"}'
            )
            RESTClient(key_file=key_file_object)
        except Exception as e:
            self.fail(f"An unexpected exception occurred: {e}")

    def test_key_file_no_key(self):
        with self.assertRaises(Exception):
            key_file_object = StringIO('{"field_1": "value_1","field_2": "value_2"}')
            RESTClient(key_file=key_file_object)

    def test_key_file_multiple_key_inputs(self):
        with self.assertRaises(Exception):
            key_file_object = StringIO('{"field_1": "value_1","field_2": "value_2"}')
            RESTClient(
                api_key=TEST_API_KEY,
                api_secret=TEST_API_SECRET,
                key_file=key_file_object,
            )

    def test_key_file_invalid_json(self):
        with self.assertRaises(Exception):
            key_file_object = StringIO(
                '"name": "test-api-key-name","privateKey": "test-api-key-private-key"'
            )
            RESTClient(key_file=key_file_object)



import json
import os
from typing import IO, Optional, Union

import requests
from requests.exceptions import HTTPError

from coinbase import jwt_generator
from coinbase.__version__ import __version__
from coinbase.constants import API_ENV_KEY, API_SECRET_ENV_KEY, BASE_URL


def handle_exception(response):
    """Raises :class:`HTTPError`, if one occurred."""

    http_error_msg = ""
    reason = response.reason

    if 400 <= response.status_code < 500:
        if (
            response.status_code == 403
            and '"error_details":"Missing required scopes"' in response.text
        ):
            http_error_msg = f"{response.status_code} Client Error: Missing Required Scopes. Please verify your API keys include the necessary permissions."
        else:
            http_error_msg = (
                f"{response.status_code} Client Error: {reason} {response.text}"
            )
    elif 500 <= response.status_code < 600:
        http_error_msg = (
            f"{response.status_code} Server Error: {reason} {response.text}"
        )

    if http_error_msg:
        raise HTTPError(http_error_msg, response=response)


class RESTBase(object):
    def __init__(
        self,
        api_key: Optional[str] = os.getenv(API_ENV_KEY),
        api_secret: Optional[str] = os.getenv(API_SECRET_ENV_KEY),
        key_file: Optional[Union[IO, str]] = None,
        base_url=BASE_URL,
        timeout=None,
    ):
        if (api_key is not None or api_secret is not None) and key_file is not None:
            raise Exception(f"Cannot specify both api_key and key_file in constructor")

        if key_file is not None:
            try:
                if isinstance(key_file, str):
                    with open(key_file, "r") as file:
                        key_json = json.load(file)
                else:
                    key_json = json.load(key_file)
                api_key = key_json["name"]
                api_secret = key_json["privateKey"]
            except json.JSONDecodeError as e:
                raise Exception(f"Error decoding JSON: {e}")

        if api_key is None:
            raise Exception(
                f"Must specify env var COINBASE_API_KEY or pass api_key in constructor"
            )
        if api_secret is None:
            raise Exception(
                f"Must specify env var COINBASE_API_SECRET or pass api_secret in constructor"
            )
        self.api_key = api_key
        self.api_secret = bytes(api_secret, encoding="utf8").decode("unicode_escape")
        self.base_url = base_url
        self.timeout = timeout

    def get(self, url_path, params: Optional[dict] = None, **kwargs):
        params = params or {}

        if kwargs:
            params.update(kwargs)

        return self.prepare_and_send_request("GET", url_path, params, data=None)

    def post(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ):
        data = data or {}

        if kwargs:
            data.update(kwargs)

        return self.prepare_and_send_request("POST", url_path, params, data)

    def put(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ):
        data = data or {}

        if kwargs:
            data.update(kwargs)

        return self.prepare_and_send_request("PUT", url_path, params, data)

    def delete(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ):
        data = data or {}

        if kwargs:
            data.update(kwargs)

        return self.prepare_and_send_request("DELETE", url_path, params, data)

    def prepare_and_send_request(
        self,
        http_method,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
    ):
        headers = self.set_headers(http_method, url_path)

        if params is not None:
            params = {key: value for key, value in params.items() if value is not None}

        if data is not None:
            data = {key: value for key, value in data.items() if value is not None}

        return self.send_request(http_method, url_path, params, headers, data=data)

    def send_request(self, http_method, url_path, params, headers, data=None):
        if data is None:
            data = {}

        url = f"https://{self.base_url}{url_path}"

        response = requests.request(
            http_method,
            url,
            params=params,
            json=data,
            headers=headers,
            timeout=self.timeout,
        )
        handle_exception(response)  # Raise an HTTPError for bad responses

        return response.json()

    def set_headers(self, method, path):
        uri = f"{method} {self.base_url}{path}"
        jwt = jwt_generator.build_rest_jwt(uri, self.api_key, self.api_secret)
        return {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + jwt,
            "User-Agent": "coinbase-advanced-py/" + __version__,
        }



from typing import Optional

from coinbase.constants import API_PREFIX


def get_candles(
    self, product_id: str, start: str, end: str, granularity: str, **kwargs
):
    """
    Get rates for a single product by product ID, grouped in buckets.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_getcandles
    """
    endpoint = f"{API_PREFIX}/products/{product_id}/candles"

    params = {
        "start": start,
        "end": end,
        "granularity": granularity,
    }

    return self.get(endpoint, params=params, **kwargs)


def get_market_trades(
    self,
    product_id: str,
    limit: int,
    start: Optional[str] = None,
    end: Optional[str] = None,
    **kwargs,
):
    """
    Get snapshot information, by product ID, about the last trades (ticks), best bid/ask, and 24h volume.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_getmarkettrades
    """
    endpoint = f"{API_PREFIX}/products/{product_id}/ticker"

    params = {"limit": limit, "start": start, "end": end}

    return self.get(endpoint, params=params, **kwargs)



from typing import List, Optional

from coinbase.constants import API_PREFIX


def create_order(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    order_configuration,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Create an order with a specified product_id (asset-pair), side (buy/sell), etc.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    endpoint = f"{API_PREFIX}/orders"

    data = {
        "client_order_id": client_order_id,
        "product_id": product_id,
        "side": side,
        "order_configuration": order_configuration,
        "self_trade_prevention_id": self_trade_prevention_id,
        "leverage": leverage,
        "margin_type": margin_type,
        "retail_portfolio_id": retail_portfolio_id,
    }

    return self.post(endpoint, data=data, **kwargs)


# Market orders
def market_order(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    quote_size: Optional[str] = None,
    base_size: Optional[str] = None,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a market order to BUY or SELL the desired product at the given market price. If you wish to purchase the
    product, provide a quote_size and if you wish to sell the product, provide a base_size.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """

    market_market_ioc = {"quote_size": quote_size, "base_size": base_size}
    filtered_market_market_ioc = {
        key: value for key, value in market_market_ioc.items() if value is not None
    }

    order_configuration = {"market_market_ioc": filtered_market_market_ioc}

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def market_order_buy(
    self,
    client_order_id: str,
    product_id: str,
    quote_size: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a market order to BUY the desired product at the given market price.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    return market_order(
        self,
        client_order_id,
        product_id,
        "BUY",
        quote_size=quote_size,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def market_order_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a market order to SELL the desired product at the given market price.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    return market_order(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Limit GTC orders
def limit_order_gtc(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a Limit Order with a GTC time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    order_configuration = {
        "limit_limit_gtc": {
            "base_size": base_size,
            "limit_price": limit_price,
            "post_only": post_only,
        }
    }

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_gtc_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a BUY Limit Order with a GTC time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    return limit_order_gtc(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        post_only=post_only,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_gtc_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a SELL Limit Order with a GTC time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    return limit_order_gtc(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        post_only=post_only,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Limit GTD orders
def limit_order_gtd(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    end_time: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a Limit Order with a GTD time-in-force policy. Unlike a Limit Order with a GTC time-in-force policy,
    this order type requires an end-time that indicates when this order should expire.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    order_configuration = {
        "limit_limit_gtd": {
            "base_size": base_size,
            "limit_price": limit_price,
            "end_time": end_time,
            "post_only": post_only,
        }
    }

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_gtd_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    end_time: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a BUY Limit Order with a GTD time-in-force policy. Unlike a Limit Order with a GTC time-in-force policy,
    this order type requires an end-time that indicates when this order should expire.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    return limit_order_gtd(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        end_time=end_time,
        post_only=post_only,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_gtd_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    end_time: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a SELL Limit Order with a GTD time-in-force policy. Unlike a Limit Order with a GTC time-in-force policy,
    this order type requires an end-time that indicates when this order should expire.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    return limit_order_gtd(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        end_time=end_time,
        post_only=post_only,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Stop-Limit GTC orders
def stop_limit_order_gtc(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a Stop Limit order with a GTC time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    order_configuration = {
        "stop_limit_stop_limit_gtc": {
            "base_size": base_size,
            "limit_price": limit_price,
            "stop_price": stop_price,
            "stop_direction": stop_direction,
        }
    }

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def stop_limit_order_gtc_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a BUY Stop Limit order with a GTC time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    return stop_limit_order_gtc(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        stop_direction=stop_direction,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def stop_limit_order_gtc_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a SELL Stop Limit order with a GTC time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    return stop_limit_order_gtc(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        stop_direction=stop_direction,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Stop-Limit GTD orders
def stop_limit_order_gtd(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    end_time: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a Stop Limit order with a GTD time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    order_configuration = {
        "stop_limit_stop_limit_gtd": {
            "base_size": base_size,
            "limit_price": limit_price,
            "stop_price": stop_price,
            "end_time": end_time,
            "stop_direction": stop_direction,
        }
    }

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def stop_limit_order_gtd_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    end_time: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a BUY Stop Limit order with a GTD time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    return stop_limit_order_gtd(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        end_time=end_time,
        stop_direction=stop_direction,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def stop_limit_order_gtd_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    end_time: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Place a SELL Stop Limit order with a GTD time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_postorder
    """
    return stop_limit_order_gtd(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        end_time=end_time,
        stop_direction=stop_direction,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def get_order(self, order_id: str, **kwargs):
    """
    Get a single order by order ID.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_gethistoricalorder
    """
    endpoint = f"{API_PREFIX}/orders/historical/{order_id}"

    return self.get(endpoint, **kwargs)


def list_orders(
    self,
    product_id: Optional[str] = None,
    order_status: Optional[List[str]] = None,
    limit: Optional[int] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    order_type: Optional[str] = None,
    order_side: Optional[str] = None,
    cursor: Optional[str] = None,
    product_type: Optional[str] = None,
    order_placement_source: Optional[str] = None,
    contract_expiry_type: Optional[str] = None,
    asset_filters: Optional[List[str]] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
):
    """
    Get a list of orders filtered by optional query parameters (product_id, order_status, etc).

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_gethistoricalorders
    """
    endpoint = f"{API_PREFIX}/orders/historical/batch"
    params = {
        "product_id": product_id,
        "order_status": order_status,
        "limit": limit,
        "start_date": start_date,
        "end_date": end_date,
        "order_type": order_type,
        "order_side": order_side,
        "cursor": cursor,
        "product_type": product_type,
        "order_placement_source": order_placement_source,
        "contract_expiry_type": contract_expiry_type,
        "asset_filters": asset_filters,
        "retail_portfolio_id": retail_portfolio_id,
    }

    return self.get(endpoint, params=params, **kwargs)


def get_fills(
    self,
    order_id: Optional[str] = None,
    product_id: Optional[str] = None,
    start_sequence_timestamp: Optional[str] = None,
    end_sequence_timestamp: Optional[str] = None,
    limit: Optional[int] = None,
    cursor: Optional[str] = None,
    **kwargs,
):
    """
    Get a list of fills filtered by optional query parameters (product_id, order_id, etc).

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_getfills
    """
    endpoint = f"{API_PREFIX}/orders/historical/fills"
    params = {
        "order_id": order_id,
        "product_id": product_id,
        "start_sequence_timestamp": start_sequence_timestamp,
        "end_sequence_timestamp": end_sequence_timestamp,
        "limit": limit,
        "cursor": cursor,
    }

    return self.get(endpoint, params=params, **kwargs)


def edit_order(
    self,
    order_id: str,
    size: Optional[str] = None,
    price: Optional[str] = None,
    **kwargs,
):
    """
    Edit an order with a specified new size, or new price. Only limit order types, with time in force type of
    good-till-cancelled can be edited.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_editorder
    """
    endpoint = f"{API_PREFIX}/orders/edit"
    data = {
        "order_id": order_id,
        "size": size,
        "price": price,
    }

    return self.post(endpoint, data=data, **kwargs)


def preview_edit_order(
    self,
    order_id: str,
    size: Optional[str] = None,
    price: Optional[str] = None,
    **kwargs,
):
    """
    Simulate an edit order request with a specified new size, or new price, to preview the result of an edit. Only
    limit order types, with time in force type of good-till-cancelled can be edited.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_previeweditorder
    """
    endpoint = f"{API_PREFIX}/orders/edit_preview"
    data = {
        "order_id": order_id,
        "size": size,
        "price": price,
    }

    return self.post(endpoint, data=data, **kwargs)


def cancel_orders(self, order_ids: List[str], **kwargs):
    """
    Initiate cancel requests for one or more orders.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_cancelorders
    """
    endpoint = f"{API_PREFIX}/orders/batch_cancel"
    data = {
        "order_ids": order_ids,
    }

    return self.post(endpoint, data=data, **kwargs)



from coinbase.constants import API_PREFIX


def get_unix_time(self, **kwargs):
    """
    Get the current time from the Coinbase Advanced API.

    https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_getunixtime
    """
    endpoint = f"{API_PREFIX}/time"

    return self.get(endpoint, **kwargs)



import secrets
import time

import jwt
from cryptography.hazmat.primitives import serialization

from coinbase.constants import BASE_URL, REST_SERVICE, WS_SERVICE


def build_jwt(key_var, secret_var, service, uri=None):
    try:
        private_key_bytes = secret_var.encode("utf-8")
        private_key = serialization.load_pem_private_key(
            private_key_bytes, password=None
        )
    except ValueError as e:
        # This handles errors like incorrect key format
        raise Exception(
            f"{e}\n"
            "Are you sure you generated your key at https://cloud.coinbase.com/access/api ?"
        )

    jwt_data = {
        "sub": key_var,
        "iss": "coinbase-cloud",
        "nbf": int(time.time()),
        "exp": int(time.time()) + 60,
        "aud": [service],
    }

    if uri:
        jwt_data["uri"] = uri

    jwt_token = jwt.encode(
        jwt_data,
        private_key,
        algorithm="ES256",
        headers={"kid": key_var, "nonce": secrets.token_hex()},
    )

    return jwt_token


def build_rest_jwt(uri, key_var, secret_var):
    return build_jwt(key_var, secret_var, REST_SERVICE, uri=uri)


def build_ws_jwt(key_var, secret_var):
    return build_jwt(key_var, secret_var, WS_SERVICE)


def format_jwt_uri(method, path):
    return f"{method} {BASE_URL}{path}"
